"use strict";  // Use ES5 strict mode for this file

const EXPORTED_SYMBOLS = ["ipdb"];  // Only symbol to be exported on Components.utils.import() for this file

Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

const XMLHttpRequest = Components.Constructor("@mozilla.org/xmlextras/xmlhttprequest;1", "nsIXMLHttpRequest");  // For loading files from chrome:// URIs

/*  NOTE:
    The IP address location database files are in the format:
        [range start integer 1][range start integer 2]...  [2 byte country code 1][2 byte country code 2]...

    For IPv4 the full 32-bit address is used. For IPv6 the 48-bit prefix of the 128-bit address is used. (cannot use full 64-bit prefix due to JS limitations)
    Each integer is big-endian binary packed -> 4 bytes for full IPv4 address, 6 bytes for IPv6 address prefix
    Each file is the concatenation of the complete ordered list of start IPs/prefixes with the corresponding list of country codes.
    All IPs/prefixes must be included with unknown/unallocated ranges listed as code "??"
*/
var IPv4DB = { type : "IPv4", filename : "ip4.db", bytesPerInt : 4 };
var IPv6DB = { type : "IPv6", filename : "ip6.db", bytesPerInt : 6 };

// Each IPDB file is loaded automatically on its first use
const pathToIPDBFiles = "chrome://flagfox/content/ipdb/";
XPCOMUtils.defineLazyGetter(this, "IPDBmetadata", function() {
    let scope = {};
    Services.scriptloader.loadSubScript(pathToIPDBFiles+"metadata.js",scope);
    return scope.IPDBmetadata;
});

var ipdb =
{
    close : function()
    {
        closeIPDBfile(IPv4DB);
        closeIPDBfile(IPv6DB);
    },

    get version()  // Generates an IPDB version/date string in the form of YYYY-M
    {
        if (!IPDBmetadata || !IPDBmetadata.created)
            return "ERROR";
        var date = new Date(IPDBmetadata.created);
        return date.getUTCFullYear() + "-" + (date.getUTCMonth()+1);  // JS months are 0-11, just to be confusing
    },

    get daysOld()
    {
        if (!IPDBmetadata || !IPDBmetadata.created)
            return Infinity;
        return (Date.now() - IPDBmetadata.created) / 86400000;
    },

    lookupIP : function(ipString)
    {
        if (ipString == "")
            return null;

        // IPv6 uses colons and IPv4 uses dots
        if (ipString.indexOf(':') == -1)
            return searchDB( IPv4DB, IPv4StringToInteger(ipString) );  // Look up normal IPv4 address

        if (ipString == "::1")  // IPv6 Localhost (prefix is zero, so can't use IPv6 prefix DB)
            return "-L";

        if (ipString.indexOf('.') != -1)  // IPv4 address embedded in an IPv6 address using mixed notation ("::ffff:" or "::" followed by standard IPv4 notation)
            return searchDB( IPv4DB, IPv4StringToInteger(ipString.substr(ipString.lastIndexOf(':')+1)) );

        var longIPv6String = expandIPv6String(ipString);  // Full IPv6 notation in use; expand all shorthand to full 32 char hex string

        for (let rule of IPv4inIPv6rules)  // Look for tunneling and embedded IPv4 to IPv6 address types
            if (longIPv6String.startsWith(rule.prefix))
                return searchDB( IPv4DB, rule.extractIPv4Integer(longIPv6String) );

        return searchDB( IPv6DB, hexStringToInteger(longIPv6String.substr(0,12)) );  // Look up normal IPv6 address prefix (48 bits = 6 bytes = 12 hex chars)
    }
};

IPv4DB.readInt = function(offset)
{
    return this.dataview.getUint32(offset);  // 4 bytes per int
};

IPv6DB.readInt = function(offset)
{
    return read48(this.dataview,offset);  // 6 bytes per int
};

/*  NOTE:
    The IPDB files are not loaded automatically on startup. Instead, each file's loading is deferred until its first usage.
    This load happens via the asynchronous callback from the DNS and proxy service handlers which calls ipdb.lookupIP().
    The loading of the file itself cannot be done fully asynchronously due to the possibility of another lookup during load.
    In order to load the IPDB file(s) asynchronously I would need to preload them earlier (or hack in some way of buffering requests).
    However, with deferred loading these files never need to be loaded unless they are used, so in this case I prefer simply deferring.
    With an extractionless install loading files from the chrome:// URI is quite fast as the XPI containing them is already loaded.
*/
function autoLoadIPDB(db)
{
/*  db.loadState values:
        undefined = not loaded
        0 = just started load
        1 = metadata is loaded (likely auto-loaded earlier)
        2 = file data is loaded
        3 = data is valid and view is loaded (ready to use)
        -1 = an error occured; do not auto-load again
*/
    if (db.loadState === undefined)
        loadIPDBfile(db);
}

function loadIPDBfile(db, async=false)
{
    if (db.data != undefined || db.loadState > 0)
        throw "Tried to load " + db.type + " DB twice!";

    db.loadState = 0;

    if (!IPDBmetadata)
        throw "IPDB metadata file failed to load";

    db.loadState = 1;

    // Open an XMLHttpRequest to the chrome:// URI and get a typed JavaScript array buffer with the data
    var request = new XMLHttpRequest();
    db.path = pathToIPDBFiles + db.filename;
    request.open("GET", db.path, async);
    request.responseType = "arraybuffer";
    request.onerror = function(event) {
        handleIPDBLoadError(db, "Flagfox ERROR attempting to load " + db.type + " DB file from: " + db.path);
    };
    request.onload = function(event)
    {
        if (db.loadState != 1)  // If closed while opening, then abort
            return;
        try {
            db.loadState = 2;
            db.data = request.response;
            loadIPDBdata(db);
        } catch (e) {
            handleIPDBLoadError(db, "Flagfox ERROR attempting to load " + db.type + " DB data: " + e);
        }
    };
    request.send();
}

function handleIPDBLoadError(db,msg)
{
    closeIPDBfile(db);
    db.loadState = -1;
    Components.utils.reportError(msg);
}

function loadIPDBdata(db)
{
    if (!db.data)
        throw "got null data on attempt to load file";

    if (!db.data.byteLength)
        throw "file loaded with zero bytes";

    if (db.data.byteLength != IPDBmetadata.size[db.type])
        throw "file is corrupt (got " + db.data.byteLength + " bytes but expected " + IPDBmetadata.size[db.type] + " bytes)";

    db.entryCount = db.data.byteLength / (db.bytesPerInt+2);  // +2 for country code
    db.countryCodesOffset = db.entryCount * db.bytesPerInt;

    if (!Number.isInteger(db.entryCount))
        throw "file is bad (got " + db.entryCount + " entries in " + db.data.byteLength + " bytes but metadata agreed)";

    db.dataview = new DataView(db.data);

    if (!db.dataview)
        throw "loaded " + db.data.byteLength + " bytes of file but could not get a DataView";

    db.loadState = 3;  // Done
}

function closeIPDBfile(db)
{
    db.path = undefined;
    db.data = undefined;
    db.dataview = undefined;
    db.entryCount = undefined;
    db.countryCodesOffset = undefined;
    db.loadState = undefined;
}

function IPv4StringToInteger(ipString)
{
    const octets = ipString.split(".");
    return (16777216 * parseInt(octets[0],10)) + (65536 * parseInt(octets[1],10)) + (256 * parseInt(octets[2],10)) + parseInt(octets[3],10);
}

function hexStringToInteger(string)
{
    return parseInt(string, 16);
}

function read48(dataview,offset)  // Reads 48 bits from a DataView (JS can't handle full 64-bit numbers)
{
    return (4294967296 * dataview.getUint16(offset)) + dataview.getUint32(offset+2);
}

function readCountryCode(db,index)  // Reads country code at given index
{
    var bytesStart = db.countryCodesOffset + index*2;
    var code = String.fromCharCode( db.dataview.getUint8(bytesStart), db.dataview.getUint8(bytesStart+1) );  // Country code (2 char string)
    return code == "??" ? null : code ;  // "??" is code for a gap in the list
}

function expandIPv6String(ipString)  // Expands an IPv6 shorthand string into its full long version (32 char hex string)
{
    var blocks = ipString.toLowerCase().split(':');
    for (let i=0; i<blocks.length; i++)
    {
        if (blocks[i].length == 0)  // Expand collapsed zeroes block
        {
            blocks[i] = "0000";
            while (blocks.length < 8)
                blocks.splice(i,0,"0000");
        }
        while (blocks[i].length < 4)  // Add leading zeroes as needed
            blocks[i] = "0" + blocks[i];
    }
    return blocks.join("");  // Drop ':' notation
}

const IPv4inIPv6rules =  // List of methods by which an IPv4 counterpart to an IPv6 address can be determined
[
    {   // IPv4 embedded IPv6 addresses not using mixed notation -> last 32-bits is IPv4 address
        prefix : "00000000000000000000",
        extractIPv4Integer : function(ipString) {
            var block6 = ipString.substr(20,4);
            if (block6 == "ffff" || block6 == "0000")
                return hexStringToInteger(ipString.substr(24,8));
            return null;  // IPv6 prefix is zero (reserved/special IP range)
        }
    },
    {   // "6to4" tunneling address -> next 32-bits is IPv4 address
        prefix : "2002",
        extractIPv4Integer : function(ipString) {
            return hexStringToInteger(ipString.substr(4,8));
        }
    },
    {   // "Teredo" tunneling address -> bitwise not of last 32-bits is IPv4 address
        prefix : "20010000",
        extractIPv4Integer : function(ipString) {
            return ~hexStringToInteger(ipString.substr(24,8));
        }
    }
];

function searchDB(db,int)  // Returns country code for integer in given DB, or null if not found
{
    if (int === null || isNaN(int))
        return null;

    autoLoadIPDB(db);  // Automatically load IPDB file on first use

    try {
        // Do a binary search loop to find given integer in ordered list of ranges
        var min = 0;
        var max = db.entryCount-1;
        var mid, bytesStart;
        while (min <= max)
        {
            mid = Math.floor((min + max) / 2);
            bytesStart = mid * db.bytesPerInt;
            if (int < db.readInt(bytesStart))
                max = mid - 1;
            else if (int >= db.readInt(bytesStart+db.bytesPerInt))  // The next number is the start of the next range; not part of this range
                min = mid + 1;
            else  /* range1start <= int && int < range2start */
                return readCountryCode(db,mid);
        }
    } catch (e) {}
    return null;
}

function generateQuickHash()  // Generates a quick 6 character alphanumeric hash ID from the metadata and IPv4 DB files
{
    // Load IPv4 DB if not done so already
    autoLoadIPDB(IPv4DB);
    if (!IPDBmetadata || !IPDBmetadata.created || !IPDBmetadata.size || !IPv4DB.entryCount || IPv4DB.loadState != 3)
        return "";

    const pointCount = 7;
    const hopLength = Math.floor((IPv4DB.data.byteLength-IPv4DB.bytesPerInt)/(pointCount-1));
    var hash = 0x9e3779b9;
    hash ^= (IPDBmetadata.created << 16);
    hash ^= IPDBmetadata.size.IPv4;
    hash ^= IPDBmetadata.size.IPv6;
    for (let i=0; i<pointCount; i++)
        hash ^= IPv4DB.readInt(i*hopLength);  // Read 'pointCount' number of integers from various parts of the stream and XOR them together

    return Math.abs(hash).toString(36);  // 6 char base-36 number string (possible characters are all alphanumeric: 0-9 + a-z)
}

// Automatically generate and cache IPDB quick hash on-demand
XPCOMUtils.defineLazyGetter(ipdb, "quickHash", function() {
    return generateQuickHash();
});
